#   nls/animation.py
#   This module define routines and type to record evolution of solution and store it into file.
#   (c) Daniel Bershatsky, 2016
#   See LICENSE for details

from __future__ import print_function
from time import time
from matplotlib import animation
from matplotlib.pyplot import figure
from .pumping import GaussianPumping


class AbstractAnimation(object):
    """Animation base class that contains common method of initialization and rendering video.
    """

    def __init__(self, model, frames, step=1):
        self.elapsed_time = 0.0
        self.model = model
        self.frames = frames
        self.step = step
        self.writer = animation.writers['ffmpeg'](fps=15, metadata={'title': 'Exciton-polariton condensation.'})

    def getElapsedTime(self):
        """Get time in seconds elapsed during animation rendering.
        """
        return self.elapsed_time

    def render(self, filename):
        """Perform initialization of render, set quality and size video attributes and then call template method that
        is defined in child class.
        """
        self.elapsed_time = -time()
        dpi = 100
        fig = figure(figsize=(16, 9), dpi=dpi)
        with self.writer.saving(fig, filename, dpi):
            for frame_id in xrange(self.frames + 1):
                self.renderFrame(frame_id)
                self.writer.grab_frame()
        self.elapsed_time += time()

    def renderFrame(self, frame_id):
        """It is not a public method that renders only one frame. This method should be overrided in child classes.
        """
        raise Exception('No frame to render: frameRender() override is needed!')

    def report(self):
        """Prints in standard output report about animation rendering. Namely, it prints seconds spent, number of
        frames and step size that is used in functional animation.
        """
        message = 'Elapsed in {0} seconds with {1} frames and {2} step.'
        print(message.format(self.elapsed_time, self.frames, self.step))


class IterationIncreaseAnimation(AbstractAnimation):
    """This class represents object that provide ability to render video that frames draw solution profile in different
    number of iteration which is increasing.
    """

    def __init__(self, model, frames, step=1):
        super(IterationIncreaseAnimation, self).__init__(model, frames, step)

    def renderFrame(self, frame_id):
        solution = self.model.solve(self.step)  # Fix references entanglement
        solution.setInitialSolution(solution.getSolution())
        solution.visualize()


class PumpingRadiusIncreaseAnimation(AbstractAnimation):
    """This class implements animation scenario that shows solution profile with different radius of pumping profile.
    In this case spacial pumping profile is gaussian.
    """

    def __init__(self, model, frames, step=1):
        super(PumpingRadiusIncreaseAnimation, self).__init__(model, frames, step)

    def renderFrame(self, frame_id):
        origin = frame_id * self.step
        pumping = GaussianPumping(power=3.0, x0=+origin, variation=6.84931506849) \
               + GaussianPumping(power=3.0, x0=-origin, variation=6.84931506849)
        self.model.solution.setPumping(pumping)
        solution = self.model.solve()  # Fix references entanglement
        solution.setInitialSolution(solution.getSolution())
        solution.visualize()


class PumpingPowerIncreaseAnimation(AbstractAnimation):
    """Funcational animation that is implemented by object of this type is increasing pumping power as time increase.
    Pumping model of a problem should provide ability to set pumping power.
    """

    def __init__(self, model, frames, step=0.1):
        super(PumpingPowerIncreaseAnimation, self).__init__(model, frames, step)

    def renderFrame(self, frame_id):
        self.model.solution.pumping.power = frame_id * self.step
        solution = self.model.solve()  # Fix references entanglement
        solution.visualize()


#   nls/pumping.py
#   The module defines abstract type and operation with pumping objects. Also it provides several predefined pumping
#   classes.
#   (c) Daniel Bershatsky, 2016
#   See LICENSE for details

from __future__ import print_function
from numpy import exp


class AbstractPumping(object):
    """Base class of pumping tree that define commont interface of pumping objects behavior.
    """

    def __add__(self, other):
        return OpSumPumping(self, other)

    def __sub__(self, other):
        return OpSubPumping(self, other)

    def __str__(self):
        return str(repr(self))

    def __repr__(self):
        return u'AbstractPumping'

    def __unicode__ (self):
        return repr(self);


class OpSumPumping(AbstractPumping):
    """Functor object that incapsulates two pumping functional objects and represents their sum.
    """

    def __init__(self, lhs, rhs):
        super(AbstractPumping, self).__init__()

        self.lhs = lhs
        self.rhs = rhs

    def __call__(self, x, t=None):
        return self.lhs(x, t) + self.rhs(x, t)

    def __repr__(self):
        return repr(self.lhs) + u' + ' + repr(self.rhs)


class OpSubPumping(AbstractPumping):
    """Functor object that incapsulates two pumping functional objects and represents their substract.
    """

    def __init__(self, lhs, rhs):
        super(AbstractPumping, self).__init__()

        self.lhs = lhs
        self.rhs = rhs

    def __call__(self, x, t=None):
        return self.lhs(x, t) + self.rhs(x, t)

    def __repr__(self):
        return repr(self.lhs) + u' - ' + repr(self.rhs)


class GaussianPumping(AbstractPumping):
    """Steady state gaussian pumping with given origin, maximum power, and decay.
    """

    def __init__(self, power=1.0, x0=0.0, y0=0.0, variation=5.0):
        super(AbstractPumping, self).__init__()

        self.power = power
        self.x0 = x0
        self.y0 = y0
        self.variation = variation

    def __call__(self, x, y=None, t=None):
        y = 0.0 if y is None else y
        return self.power * exp( - ((x - self.x0) ** 2 + (y - self.y0) ** 2) / (2.0 * self.variation))

    def __repr__(self):
        return u'{0} exp(-{1} (x - {2})^2)'.format(self.power, 1.0 / (2.0 * self.variation), self.x0)



#   nls/__init__.py
#   (c) Daniel Bershatsky, 2016
#   See LICENSE for details

from __future__ import print_function

from .animation import *
from .model import *
from .pumping import *

__all__ = [
    'animation',
    'model',
    'native'
    'pumping',
    'solver',
]

MAJOR = 0
MINOR = 1
PATCH = 0


def version():
    return  MAJOR, MINOR, PATCH


#!/usr/bin/env python2
#
#   (c) Daniel Bershatsky <daniel.bershatsky@skolkovotech.ru>, 2015
#

from __future__ import print_function
from numpy import arange
from matplotlib.pyplot import plot, show, legend
from sys import argv


def visualize(filename='solution'):
    with open(filename) as f:
        u = [float(x) for x in f.read().split()]
    
    n, h = len(u), 0.001
    plot(arange(0.0, n * h, h), u, label=filename)

def main():
    visualize(argv[1] if len(argv) == 2 else 'solution')
    legend(loc='best')
    show()


if __name__ == '__main__':
    main()


#   nls/model.py
#   This module define core abstractions that maps to problem and model definition.
#   (c) Daniel Bershatsky, 2016
#   See LICENSE for details

from __future__ import print_function

from pprint import pprint
from time import time
from datetime import datetime
from numpy import array, exp, arange, ones, zeros, meshgrid, mgrid, linspace
from scipy.io import loadmat, savemat
from mpl_toolkits.mplot3d import Axes3D
from matplotlib import animation
from matplotlib.pyplot import figure, plot, show, title, xlabel, ylabel, subplot, legend, xlim, ylim, contourf, hold

from .animation import *
from .native import *
from .pumping import *
from .solver import *


class Problem(object):
    """Entry point in any computation. It implements design pattern `Factory` that used to construct object of type
    `Model`.
    """

    def model(self, *args, **kwargs):
        """
        Piority of Arguments: Arguments passed in `kwargs` has the most piority, 'param' key in `kwargs` has less
        piority than `kwargs` and dictionary arguments in `args` have the least piority. Other arguments are ignored.
        Argument List:
            model - set model type, default value 'default';
            dx - default value '1.0e-1';
            dt - default value '1.0e-3';
            t0 - default value '0.0';
            u0 - default value '1.0e-1';
            order - default value '5';
            pumping - default value ``;
            !original_params - default value `{}`;
            !dimless_params - default value `{}`;
        """
        if 'params' in kwargs:
            params = kwargs.pop('params')

        kwargs['model'] == 'default' if 'model' not in kwargs else kwargs['model']

        if 'model' in kwargs and kwargs['model'] in ('1d', 'default'):
            return self.fabricateModel1D(*args, **kwargs)
        elif 'model' in kwargs and kwargs['model'] == '2d':
            return self.fabricateModel2D(*args, **kwargs)
        else:
            raise Exception('Unknown model passed!')

    def fabricateModel1D(*args, **kwargs):
        kwargs['dx'] = 1.0e-1 if 'dx' not in kwargs else kwargs['dx']
        kwargs['dt'] = 1.0e-3 if 'dt' not in kwargs else kwargs['dt']
        kwargs['t0'] = 0.0e+0 if 't0' not in kwargs else kwargs['t0']
        kwargs['u0'] = 1.0e-1 if 'u0' not in kwargs else kwargs['u0']
        kwargs['order'] = 5 if 'order' not in kwargs else kwargs['order']
        kwargs['pumping'] = GaussianPumping() if 'pumping' not in kwargs else kwargs['pumping']
        kwargs['num_nodes'] = 1000 if 'num_nodes' not in kwargs else kwargs['num_nodes']
        kwargs['num_iters'] = 100000 if 'num_iters' not in kwargs else kwargs['num_iters']

        if type(kwargs['u0']) in (int, float, complex):
            kwargs['u0'] = kwargs['u0'] * ones(kwargs['num_nodes']) 

        return Model1D(**kwargs)

    def fabricateModel2D(self, *args, **kwargs):
        kwargs['dx'] = 1.0e-1 if 'dx' not in kwargs else kwargs['dx']
        kwargs['dt'] = 1.0e-3 if 'dt' not in kwargs else kwargs['dt']
        kwargs['t0'] = 0.0e+0 if 't0' not in kwargs else kwargs['t0']
        kwargs['u0'] = 1.0e-1 if 'u0' not in kwargs else kwargs['u0']
        kwargs['order'] = 3 if 'order' not in kwargs else kwargs['order']
        kwargs['pumping'] = GaussianPumping() if 'pumping' not in kwargs else kwargs['pumping']
        kwargs['num_nodes'] = 40 if 'num_nodes' not in kwargs else kwargs['num_nodes']
        kwargs['num_iters'] = 1000 if 'num_iters' not in kwargs else kwargs['num_iters']

        if type(kwargs['u0']) in (int, float, complex):
            kwargs['u0'] = kwargs['u0'] * ones((kwargs['num_nodes'], kwargs['num_nodes'])) 

        return Model2D(**kwargs)


class AbstractModel(object):
    """Base type for objects which constructed with `Problem` class. Child object of this class implements computation
    and other related routines. This class defines common routines of initialization, solving, and model storage.
    """

    def __init__(self, *args, **kwargs):
        pprint({
            'dt': kwargs['dt'],
            'dx': kwargs['dx'],
            'order': kwargs['order'],
            'num_nodes': kwargs['num_nodes'],
            'num_iters': kwargs['num_iters'],
            'pumping': kwargs['pumping'],
            'originals': kwargs['original_params'],
            })
        self.solution = Solution(kwargs['dt'], kwargs['dx'], kwargs['num_nodes'], kwargs['order'], kwargs['num_iters'],
                             kwargs['pumping'], kwargs['original_params'], kwargs['u0'])
        self.solver = None

    def solve(self, num_iters=None):
        """Call solver that is aggregated certain child objects.
        """
        return self.solver(num_iters)

    def store(self, filename=None, label='', desc='', date=datetime.now()):
        """Store object to mat-file. TODO: determine format specification
        """
        filename = filename if filename else str(date).replace(' ', '_') + '.mat'

        matfile = {}
        matfile['desc'] = desc
        matfile['label'] = label
        matfile['originals'] = {}

        savemat(filename, matfile)

    def restore(self, filename):
        """Restore object from mat-file. TODO: determine format specification
        """
        matfile = loadmat(filename)

        self.desc = str(matfile['desc'][0]) if matfile['desc'].size else ''
        self.label = str(matfile['label'][0]) if matfile['label'].size else ''
        self.originals = {}

        return self


class Model1D(AbstractModel):
    """Default model that is NLS equation with reservoir in axe symmentic case.
    """

    def __init__(self, *args, **kwargs):
        super(Model1D, self).__init__(*args, **kwargs)

        self.solver = Solver1D(self.solution)


class Model2D(AbstractModel):
    """Model that is NLS equation with reservoir on two dimensional grid.
    """

    def __init__(self, *args, **kwargs):
        super(Model2D, self).__init__(*args, **kwargs)

        self.solver = Solver2D(self.solution)


class Solution(object):
    """Object that represents solution of a given model. Also it contains all model parameters and has ability to store
    and to load solution. TODO: improve design.
    """

    def __init__(self, dt, dx, num_nodes, order, num_iters, pumping, originals, init_solution):
        self.dt = dt
        self.dx = dx
        self.order = order
        self.num_nodes = num_nodes
        self.num_iters = num_iters
        self.pumping = pumping
        self.init_sol = init_solution
        self.solution = None
        self.originals = originals
        self.coeffs = zeros(23)
        self.elapsed_time = 0.0

        hbar = 6.61e-34
        m_e = 9.1e-31
        m_0 = 1.0e-5 * m_e

        t0 = 2.0 / originals['gamma']
        x0 = (hbar * t0 / 2 / m_0) ** 0.5
        n0 = 2.0 / (originals['R'] * t0)
        phi0 = t0

        # NLS equation coeficients
        self.coeffs[0] = 1.0  # \partial_t
        self.coeffs[1] = 1.0  # \nabla^2
        self.coeffs[2] = 1.0  #
        self.coeffs[3] = 1.0  # linear damping
        self.coeffs[4] = originals['g'] * phi0 ** 3  # nonlinearity
        self.coeffs[5] = originals['tilde_g'] * phi0 * n0  # interaction to reservoir

        # Reservoir equation coefficients
        self.coeffs[10] = 0.0  # \parital_t
        self.coeffs[11] = 1.0 / (n0 * originals['gamma_R'])  # pumping coefficient
        self.coeffs[12] = 1.0  # damping
        self.coeffs[13] = originals['R'] * phi0 ** 2 / originals['gamma_R']  # interaction term
        self.coeffs[14] = 0.0  # diffusive term

        print(self.coeffs)

    def getTimeStep(self):
        return self.dt

    def getSpatialStep(self):
        return self.dx

    def getApproximationOrder(self):
        return self.order

    def getNumberOfNodes(self):
        return self.num_nodes

    def getNumberOfIterations(self):
        return self.num_iters

    def getPumping(self):
        if len(self.init_sol.shape) == 1:
            right = self.num_nodes * self.dx
            left = 0.0
            x = linspace(left, right, self.num_nodes)
            grid = meshgrid(x)
            return self.pumping(*grid)
        else:
            right = self.num_nodes * self.dx / 2
            left = -right
            x = linspace(left, right, self.num_nodes)
            grid = meshgrid(x, x)
            return self.pumping(*grid)

    def getCoefficients(self):
        return self.coeffs

    def getInitialSolution(self):
        return self.init_sol

    def getSolution(self):
        return self.solution

    def getElapsedTime(self):
        return self.elapsed_time

    def setNumberOfIterations(self, num_iters):
        self.num_iters = num_iters

    def setPumping(self, pumping):
        self.pumping = pumping

    def setInitialSolution(self, solution):
        self.init_sol = solution

    def setSolution(self, solution):
        self.solution = solution

    def setElapsedTime(self, seconds):
        self.elapsed_time = seconds

    def visualize(self):
        if len(self.init_sol.shape) == 1:
            self.visualize1d()
        else:
            self.visualize2d()

    def visualize1d(self):
        x = arange(0.0, self.dx * self.num_nodes, self.dx)
        p = self.pumping(x)  # pumping profile
        u = (self.solution.conj() * self.solution).real  # density profile
        n = self.coeffs[11] *  p / (self.coeffs[12] + self.coeffs[13] * u)

        def rect_plot(subplot_number, value, label, name, labelx, labely, xmax=20):
            subplot(2, 3, subplot_number)
            hold(False)
            plot(x, value, label=label)
            xlim((0, xmax))
            legend(loc='best')
            title(name)
            xlabel(labelx)
            ylabel(labely)

        rect_plot(1, p, 'pumping', 'Pumping profile.', 'r', 'p')
        rect_plot(2, u, 'density', 'Density distribution of BEC.', 'r', 'u')
        rect_plot(3, n, 'reservoir', 'Density distribution of reservoir.', 'r', 'n')

        def polar_plot(subplot_number, value, xmax=20):
            hold(False)
            subplot(2, 3, subplot_number, polar=True)
            theta = arange(0, 2 * 3.14 + 0.1, 0.1)
            contourf(theta, x, array([value for _ in theta]).T)
            ylim((0, xmax))

        polar_plot(4, p)
        polar_plot(5, u)
        polar_plot(6, n)

    def visualize2d(self):
        right = self.num_nodes * self.dx / 2
        left = -right
        x = linspace(left, right, self.num_nodes)
        gx, gy = meshgrid(x, x)
        p = self.getPumping()
        u = (self.solution.conj() * self.solution).real  # density profile
        n = self.coeffs[11] *  p / (self.coeffs[12] + self.coeffs[13] * u)

        fig = figure()

        def surface_plot(subplot_number, value, label, name, labels):
            ax = fig.add_subplot(130 + subplot_number, projection='3d')
            ax.plot_surface(gx, gy, value, label=label)
            ax.set_xlabel('x')
            ax.set_ylabel('y')
            ax.set_title(name)

        surface_plot(1, p, 'pumping', 'Pumping profile.', ('x', 'y', 'p'))
        surface_plot(2, u, 'density', 'Density distribution of BEC.', ('x', 'y', 'u'))
        surface_plot(3, n, 'reservoir', 'Density distribution of reservoir.', ('x', 'y', 'n'))

    def show(self):
        show()

    def store(self, filename=None, label='', desc='', date=datetime.now()):
        """Store object to mat-file. TODO: determine format specification
        """
        filename = filename if filename else str(date).replace(' ', '_') + '.mat'

        matfile = {}
        matfile['desc'] = desc
        matfile['dimlesses'] = self.coeffs
        matfile['elapsed_time'] = self.elapsed_time
        matfile['label'] = label
        matfile['originals'] = self.originals
        matfile['pumping'] = self.getPumping()
        matfile['solution'] = self.solution

        savemat(filename, matfile)

    def restore(self, filename):
        """Restore object from mat-file. TODO: determine format specification
        """
        matfile = loadmat(filename)

        self.desc = str(matfile['desc'][0]) if matfile['desc'].size else ''
        self.coeffs = matfile['dimlesses'].T
        self.elapsed_time = matfile['elapsed_time']
        self.label = str(matfile['label'][0]) if matfile['label'].size else ''
        self.originals = {}
        self.solution = matfile['solution'].T

        return self

    def report(self):
        message = 'Elapsed in {0} seconds with {1} iteration on {2} grid nodes.'
        print(message.format(self.elapsed_time, self.num_iters, self.num_nodes))


#   nls/solver.py
#   This file contains solver abstraction layer that hides relevant implimentation.
#   (c) Daniel Bershatsky, 2016
#   See LICENSE for details

from __future__ import print_function
from .native import nls
from .model import *


class AbstractSolver(object):
    """Wrapper(or Facade template) of native fortran subroutine `solve_nls`.
    """
    
    def __init__(self, solution):
        self.solution = solution
        self.elapsed_time = 0.0

    def __call__(self, num_iters=None):
        if num_iters:
            self.solution.setNumberOfIterations(num_iters)

        self.elapsed_time = -time()
        self.solution.setSolution(self.solve(
            self.solution.getTimeStep(),
            self.solution.getSpatialStep(),
            self.solution.getApproximationOrder(),
            self.solution.getNumberOfIterations(),
            self.solution.getPumping(),
            self.solution.getCoefficients(),
            self.solution.getInitialSolution()))
        self.elapsed_time += time()
        self.solution.setElapsedTime(self.elapsed_time)

        return self.solution

    def solve(*argv, **kwargs):
        """This method should be override in child classes that specify solver routine(template method pattern).
        """
        raise Exception('AbstractSolver: native solver routine is not passed!')


class Solver1D(AbstractSolver):
    """One dimensional solver that call native Fortran routine that solves NLS equation in axial symmentry in 2D.
    """

    def __init__(self, solution):
        super(Solver1D, self).__init__(solution)

    def solve(self, *args, **kwargs):
        return nls.solve_nls(*args, **kwargs)


class Solver2D(AbstractSolver):
    """One dimensional solver that call native Fortran routine that solves NLS equation on a squared grid.
    """


    def __init__(self, solution):
        super(Solver2D, self).__init__(solution)

    def solve(self, *args, **kwargs):
        return nls.solve_nls_2d(*args, **kwargs)


#!/usr/bin/env python2
#
#   In this file test glue layer between Fortran 95 code and module generated by f2py from numpy.
#   Here and bellow `native solution` means solution that was obtained with native fortran code
#   and in opposite `glue solution` means native solution transmitted through glue layer.
#
#   (c) Daniel Bershatsky <daniel.bershatsky@skolkovotech.ru>, 2016
#

from __future__ import print_function
from nls import nls
from numpy import array, arange, diag, abs, max
from scipy.sparse import diags
from subprocess import Popen, PIPE


VERBOSE = True
FLOAT_THRESHOLD = 1e-3

NX, ITERS, ORDER = 200, 10000, 5
DT, DX = 1.0e-6, 1.0e-3


def test_make_banded_matrix():
    print('Test make_banded_matrix():')

def test_banded_matvec():
    print('Test banded_matvec():')

def test_make_laplacian():
    print('Test make_laplacian():')
    n, m, h = 8, 8, 0.1
    r = arange(0, h * m - 1e-10, h)
    D1 = diags((1, -8, 0, 8, -1), (-2, -1, 0, 1, 2), shape=(m, m)).toarray()
    D2 = diags((-1, 16, -30, 16, -1), (-2, -1, 0, 1, 2), shape=(m, m)).toarray()
    r[0] = 1
    D1[0, :] = 0
    D1[1, 1] += 1
    D2[0, :3] = [-60, 64, -4]
    D2[1, 1] += -1
    D = D2 / (24 * h ** 2) + diag(1.0 / r).dot(D1)  / (12 * h)
    L = nls.make_laplacian(n, 5, h)
    tolerance = 5.0e-5
    for k in xrange(0, 3):
        err = max(abs(diag(D, 2 - k) - L[k, 2 - k:]))
        print('Diagonal # ', 2 - k, ':    ', err)
        if err >= tolerance:
            print(diag(D, 2 - k))
            print(L[k, 2 - k:])
    for k in xrange(1, 3):
        err = max(abs(diag(D, -k) - L[2 + k, :-k]))
        print('Diagonal #', -k, ':    ', err)
        if err >= tolerance:
            print(diag(D, -k))
            print(L[2 + k, :-k])

def test_runge_kutta():
    """Call runge_kutta() fortran subroutine directly through f2py wrapper layer.
    """
    print('Test runge_kutta():')
    u0 = array([1, 1, 1, 1, 1, 1, 1, 1])
    op = nls.make_laplacian(8, 5, 0.01)
    u = nls.runge_kutta(0.00001, 0.0, 0.01, u0, op, 10)
    print(u)

def get_native_solution():
    pid = Popen(['./solve'], stdin=PIPE, stdout=PIPE, stderr=PIPE)
    output, _ = pid.communicate()
    return array([float(x) for x in output.split()])


def get_glue_solution():
    u = nls.solve_nls(DT, DX, NX, ORDER, ITERS)
    return abs(u.conj() * u)

def compare_solutions():
    native = get_native_solution()
    glue = get_native_solution()
    difference = abs(native - glue) < FLOAT_THRESHOLD
    different = reduce(lambda acc, val: acc if val[1] else acc + [val[0]], enumerate(difference), [])
    return all(difference), different

def report_solution_comparison(is_equal, different):
    print('Equal:', is_equal)
    
    if VERBOSE and not is_equal:
        print('Next elements are different:')
        print(different)
        print('Native:\n', native[different])
        print('Glue:\n', glue[different])

def test_glue_layer():
    """Run native fortran code and its python bindings in order to compare these solutions and test
    their equality.
    """
    print('Test glue layer:')
    is_equal, different = compare_solutions()
    report_solution_comparison(is_equal, different)

def test():
    test_make_banded_matrix()
    test_banded_matvec()
    test_make_laplacian()
    test_runge_kutta()
    test_glue_layer()


if __name__ == '__main__':
    test()
